// SPDX-License-Identifier: GPL-2.0

#include "gcc-common.h"
#include "gcc-ksyms.h"

__visible int plugin_is_GPL_compatible;

static unsigned int sp_mask, canary_offset;

static unsigned int arm_pertask_ssp_rtl_execute(void)
{
	rtx_insn *insn;

	/* resolve enums lazy, as plugins are in use even before ksyms.h is generated */
	/* TODO: move to a "late" plugin infrastructure to get rid of this */
	if (!sp_mask) {
		if (THREAD_STACK_MASK != NULL_TREE && TI_STACK_CANARY != NULL_TREE) {
			canary_offset = tree_to_shwi(TI_STACK_CANARY);
			sp_mask = tree_to_shwi(THREAD_STACK_MASK);
		}
	}

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		const char *sym;
		rtx body;
		rtx mask, masked_sp;

		/*
		 * Find a SET insn involving a SYMBOL_REF to __stack_chk_guard
		 */
		if (!INSN_P(insn))
			continue;
		body = PATTERN(insn);
		if (GET_CODE(body) != SET ||
		    GET_CODE(SET_SRC(body)) != SYMBOL_REF)
			continue;
		sym = XSTR(SET_SRC(body), 0);
		if (strcmp(sym, "__stack_chk_guard"))
			continue;

		/* ensure it's really the one we're looking for */
		gcc_assert(SYMBOL_REF_DECL(SET_SRC(body)) == default_stack_protect_guard());

		/* now we need it for real! */
		if (!sp_mask) {
			error_at(INSN_LOCATION(insn), "missing value for 'THREAD_STACK_MASK', cannot instrument %qE", current_function_decl);
			gcc_unreachable();
		}

		/*
		 * Replace the source of the SET insn with an expression that
		 * produces the address of the copy of the stack canary value
		 * stored in struct thread_info
		 */
		mask = GEN_INT(sext_hwi(sp_mask, GET_MODE_PRECISION(Pmode)));
		masked_sp = gen_reg_rtx(Pmode);

		emit_insn_before(gen_rtx_set(masked_sp,
					     gen_rtx_AND(Pmode,
							 stack_pointer_rtx,
							 mask)),
				 insn);

		SET_SRC(body) = gen_rtx_PLUS(Pmode, masked_sp,
					     GEN_INT(canary_offset));

		df_insn_rescan(insn);
	}
	return 0;
}

#define PASS_NAME arm_pertask_ssp_rtl
#define NO_GATE
#if BUILDING_GCC_VERSION <= 4007
#define TODO_FLAGS_FINISH TODO_dump_func | TODO_df_verify
#else
#define TODO_FLAGS_FINISH TODO_dump_func | TODO_verify_rtl_sharing | TODO_df_verify
#endif
#include "gcc-generate-rtl-pass.h"

#if BUILDING_GCC_VERSION >= 9000
static bool no(void)
{
	return false;
}

static void arm_pertask_ssp_start_unit(void *gcc_data, void *user_data)
{
	targetm.have_stack_protect_combined_set = no;
	targetm.have_stack_protect_combined_test = no;
}
#endif

__visible int plugin_init(struct plugin_name_args *plugin_info,
			  struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument *argv = plugin_info->argv;
	int i;

	PASS_INFO(arm_pertask_ssp_rtl, "expand", 1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	/* no options supported */
	for (i = 0; i < argc; ++i)
		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &arm_pertask_ssp_rtl_pass_info);
#if BUILDING_GCC_VERSION >= 9000
	register_callback(plugin_name, PLUGIN_START_UNIT, arm_pertask_ssp_start_unit, NULL);
#endif

	return 0;
}
