/*
 * Copyright 2019-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to convert k*alloc calls into type specific slab allocations
 *
 * TODO:
 * - too many to list
 *
 * BUGS:
 * - none known
 */

#include "autoslab.h"

__visible int plugin_is_GPL_compatible;

static struct plugin_info autoslab_plugin_info = {
	.version	= "201912090900",
	.help		= "verbose\tlog autoslab decisions\n"
};

bool verbose;

static void find_autoslab_info_type(void *event_data __unused, void *data __unused)
{
	tree type = (tree)event_data;
	const_tree name;

	if (type == NULL_TREE || type == error_mark_node)
		return;

	if (TREE_CODE(type) != RECORD_TYPE )
		return;

	if (TYPE_FIELDS(type) == NULL_TREE)
		return;

	name = type_name(type);
	if (name == NULL_TREE)
		return;

	if (!id_equal(name, "autoslab_info"))
		return;

	if (autoslab_info_type) {
		gcc_assert(autoslab_info_type == type);
		return;
	}

	autoslab_info_type = type;
	build_pointer_type(type);
}

void autoslab_find_decls(void *event_data, void *data __unused)
{
	tree decl = (tree)event_data;
	enum kmem_call_names kmem_call;

	if (decl == error_mark_node)
		return;

	switch (TREE_CODE(decl)) {
	default:
		return;

	case FUNCTION_DECL:
		if (DECL_ARTIFICIAL(decl))
			return;

		if (DECL_ABSTRACT_ORIGIN(decl) && DECL_ABSTRACT_ORIGIN(decl) != decl)
			return;

		gcc_assert(DECL_NAME_POINTER(decl)[0]);

		for (kmem_call = KMEM_CACHE_ALLOC; kmem_call < KMEM_CACHE_NUM; ++kmem_call) {
			if (!id_equal(DECL_NAME(decl), kmem_cache_decls[kmem_call].name))
				continue;

			if (kmem_cache_decls[kmem_call].decl != NULL_TREE) {
				gcc_assert(decl == kmem_cache_decls[kmem_call].decl);
				continue;
			}

			kmem_cache_decls[kmem_call].decl = decl;
			kmem_cache_decls[kmem_call].used = TREE_USED(decl);
			TREE_USED(decl) = 1;
			kmem_cache_decls[kmem_call].preserved = DECL_PRESERVE_P(decl);
			DECL_PRESERVE_P(decl) = 1;
			return;
		}
	}
}

#if BUILDING_GCC_VERSION < 4007
static void autoslab_find_all_decls(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr cnode;
	varpool_node_ptr vnode;

	FOR_EACH_FUNCTION(cnode) {
		if (cnode->thunk.thunk_p || cnode->alias)
			continue;
		if (NODE_GLOBAL(cnode)->inlined_to)
			continue;

		autoslab_find_decls(NODE_DECL(cnode), NULL);
	}
}
#endif

/*
 * undo the forced used mark on kmem_cache_*
 * the instrumentation no longer needs it as we're past inlining here
 */
static void unmark_kmem_cache_decls(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr node;
	enum kmem_call_names kmem_call;

	for (kmem_call = KMEM_CACHE_ALLOC; kmem_call < KMEM_CACHE_NUM; ++kmem_call) {
		tree fndecl = kmem_cache_decls[kmem_call].decl;

		if (!fndecl)
			continue;

		TREE_USED(fndecl) = kmem_cache_decls[kmem_call].used;
		DECL_PRESERVE_P(fndecl) = kmem_cache_decls[kmem_call].preserved;
		node = cgraph_get_node(fndecl);
		if (!node)
			continue;

#if BUILDING_GCC_VERSION <= 4007
		NODE_SYMBOL(node)->needed = kmem_cache_decls[kmem_call].preserved;
#else
		NODE_SYMBOL(node)->force_output = kmem_cache_decls[kmem_call].preserved;
#endif
	}
}

static void register_attributes(void *event_data __unused, void *data __unused)
{
	register_attribute(&alloc_flags_attr);
	register_attribute(&alloc_attr);
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	const char * const plugin_name = plugin_info->base_name;

	PASS_INFO(autoslab, "alias", 1, PASS_POS_INSERT_BEFORE);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "verbose")) {
			verbose = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &autoslab_plugin_info);
	register_callback(plugin_name, PLUGIN_FINISH_TYPE, find_autoslab_info_type, NULL);
#if BUILDING_GCC_VERSION < 4007
	register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, autoslab_find_all_decls, NULL);
#else
	register_callback(plugin_name, PLUGIN_FINISH_DECL, autoslab_find_decls, NULL);
#endif
	register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, autoslab_find_decls, NULL);
	register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, unmark_kmem_cache_decls, NULL);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &autoslab_pass_info);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_autoslab);
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	return 0;
}
