/*
 * Copyright 2019-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "autoslab.h"

GTY(()) tree autoslab_info_type;

GTY(()) struct kmem_cache_decl kmem_cache_decls[KMEM_CACHE_NUM] = {
#define INIT_KMEM_CACHE_DECLS(NAME, _name)	\
[KMEM_CACHE_##NAME] = {				\
		.name = "kmem_cache_"_name,	\
		.decl = NULL_TREE,		\
		.used = false,			\
		.preserved = false,		\
}

	INIT_KMEM_CACHE_DECLS(ALLOC,       "alloc"),
	INIT_KMEM_CACHE_DECLS(ALLOC_NODE,  "alloc_node"),

#undef INIT_KMEM_CACHE_DECLS
};

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_gt_autoslab[] = {
	{
		.base = &kmem_cache_decls[0].decl,
		.nelt = ARRAY_SIZE(kmem_cache_decls),
		.stride = sizeof(kmem_cache_decls[0]),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &autoslab_info_type,
		.nelt = 1,
		.stride = sizeof(autoslab_info_type),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	LAST_GGC_ROOT_TAB
};

static struct {
	const char *kalloc_name;
	enum kmem_call_names kmem_name;
} kalloc_functions[] = {
	{ "kmalloc",			KMEM_CACHE_ALLOC },
	{ "kmalloc_node",		KMEM_CACHE_ALLOC_NODE },
	{ "__kmalloc_track_caller",	KMEM_CACHE_ALLOC },
	{ "__kmalloc_node_track_caller",KMEM_CACHE_ALLOC_NODE },
};

static size_t is_kalloc(const char *fnname)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(kalloc_functions); i++)
		if (!strcmp(fnname, kalloc_functions[i].kalloc_name))
			break;
	return i;
}

static tree autoslab_build_var(location_t loc, tree id, tree type, tree ctor, const char *section)
{
	tree var;

	var = build_decl(loc, VAR_DECL, id, type);
	DECL_INITIAL(var) = ctor;
	DECL_CONTEXT(var) = DECL_CONTEXT(current_function_decl);
	DECL_EXTERNAL(var) = 0;
	DECL_IGNORED_P(var) = 0;
	DECL_ARTIFICIAL(var) = 1;
	DECL_PRESERVE_P(var) = 1;

	TREE_STATIC(var) = 1;
	TREE_PUBLIC(var) = 1;
	TREE_ADDRESSABLE(var) = 1;
	TREE_USED(var) = 1;

	if (section)
		set_decl_section_name(var, section);
	else
		TREE_READONLY(var) = 1;
	add_referenced_var(var);
	mark_sym_for_renaming(var);
	varpool_add_new_variable(var);
#if BUILDING_GCC_VERSION <= 4008
	if (!flag_toplevel_reorder)
		varpool_assemble_decl(varpool_node_for_decl(var));
#endif

	return var;
}

static tree __autoslab_get_alloc_type(gcall *kalloc_call_stmt, gimple_stmt_ptr use_stmt)
{
	tree lhs, lhs_type;

	switch (gimple_code(use_stmt)) {
	default:
		print_gimple_stmt(stderr, kalloc_call_stmt, 0, TDF_LINENO);
		print_gimple_stmt(stderr, use_stmt, 0, TDF_LINENO);
		gcc_unreachable();

	case GIMPLE_ASSIGN:
		lhs = gimple_assign_lhs(use_stmt);
		break;

	case GIMPLE_CALL:
		return NULL_TREE;

	case GIMPLE_COND:
		return NULL_TREE;

	case GIMPLE_PHI:
		lhs = gimple_phi_result(use_stmt);
		break;

	case GIMPLE_RETURN:
		return NULL_TREE;
	}

	lhs_type = TREE_TYPE(lhs);
	if (!POINTER_TYPE_P(lhs_type))
		return NULL_TREE;

	lhs_type = TREE_TYPE(lhs_type);
	if (!RECORD_OR_UNION_TYPE_P(lhs_type))
		return NULL_TREE;

	return lhs_type;
}

static tree autoslab_get_alloc_type(gcall *kalloc_call_stmt)
{
	tree lhs, lhs_type;
	use_operand_p use_p;
	gimple_stmt_ptr use_stmt;

	lhs = gimple_call_lhs(kalloc_call_stmt);
	gcc_assert(lhs);
	lhs_type = TREE_TYPE(lhs);
	gcc_assert(POINTER_TYPE_P(lhs_type));
	lhs_type = TREE_TYPE(lhs_type);

	if (RECORD_OR_UNION_TYPE_P(lhs_type))
		return lhs_type;

	if (single_imm_use(lhs, &use_p, &use_stmt))
		return __autoslab_get_alloc_type(kalloc_call_stmt, use_stmt);

	return NULL_TREE;
}

/*
 * if the lhs type can be slabified (for now, if it's a struct type)
 * then convert the k*alloc*(size, flags, ...) call into kmem_cache_*alloc*(cache, flags, ...)
 *
 * in particular:
 * 1. kmalloc      -> kmem_cache_alloc
 * 2. kmalloc_node -> kmem_cache_alloc_node
 * 3. kzalloc      -> kmem_cache_alloc
 * 4. kzalloc_node -> kmem_cache_alloc_node
 * 5. kcalloc, kmalloc_array -> kmem_cache_alloc
 * 6. kcalloc_node, kmalloc_array_node -> kmem_cache_alloc_node
 */
static void autoslab_convert_call(gimple_stmt_iterator *gsi,
				  gcall *kalloc_call_stmt,
				  enum kmem_call_names kmem_call,
				  tree count_arg,
				  tree size_arg,
				  tree flags_arg)
{
	tree lhs_type, size, flags;
	tree autoslab, autoslab_cache_id;
	tree kmem_cache_alloc, kmem_cache;
	unsigned HOST_WIDE_INT size_val, lhs_size_val, lhs_align_val, flags_val;
	location_t loc = gimple_expansion_point_location(kalloc_call_stmt);
	gcall *kmem_call_stmt;
	char *autoslab_cache_name, *autoslab_init_name, *autoslab_name;
	const char *lhs_type_name_ptr;
	int autoslab_cache_name_len, autoslab_init_name_len, autoslab_name_len;
	varpool_node_ptr var;

	lhs_type = autoslab_get_alloc_type(kalloc_call_stmt);
	size = gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(size_arg) - 1);
	flags = gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(flags_arg) - 1);

	if (lhs_type != NULL_TREE) {
		const_tree lhs_type_name;

		lhs_type_name = type_name(lhs_type);
		if (lhs_type_name != NULL_TREE)
			lhs_type_name_ptr = IDENTIFIER_POINTER(lhs_type_name);
		else
			lhs_type_name_ptr = "_.";
	} else
		lhs_type_name_ptr = "_.";

	if (!TREE_CONSTANT(size)) {
		if (verbose)
			inform(loc, "autoslab: no varsize %qE %s", flags, lhs_type_name_ptr);
		gcc_assert(TREE_CODE(size) == SSA_NAME);
		if (!TREE_CONSTANT(flags)) {
			if (verbose)
				inform(loc, "autoslab: no varflags/varsize %qE %s", flags, lhs_type_name_ptr);
			gcc_assert(TREE_CODE(flags) == SSA_NAME);
		}
		return;
	}

	if (integer_zerop(size)) {
		// convert 0-sized allocations to ZERO_SIZE_PTR
		gimple_stmt_ptr assign;
		tree lhs;

		lhs = gimple_call_lhs(kalloc_call_stmt);
		assign = gimple_build_assign(lhs, fold_convert_loc(loc, TREE_TYPE(lhs), ZERO_SIZE_PTR));
		unlink_stmt_vdef(kalloc_call_stmt);
#if BUILDING_GCC_VERSION >= 5000
		gcc_assert(!gsi_replace(gsi, assign, true));
#else
		gsi_replace(gsi, assign, true);
#endif

		if (verbose)
			inform(loc, "autoslab: yes nullsize %qE %s", flags, lhs_type_name_ptr);
		return;
	}

	size_val = TREE_INT_CST_LOW(size);

	if (tree_int_cst_lt(KMALLOC_MAX_SIZE, size)) {
		// convert oversized allocations to NULL
		gimple_stmt_ptr assign;
		tree lhs;

		lhs = gimple_call_lhs(kalloc_call_stmt);
		assign = gimple_build_assign(lhs, fold_convert_loc(loc, TREE_TYPE(lhs), null_pointer_node));
		unlink_stmt_vdef(kalloc_call_stmt);
#if BUILDING_GCC_VERSION >= 5000
		gcc_assert(!gsi_replace(gsi, assign, true));
#else
		gsi_replace(gsi, assign, true);
#endif

		if (verbose)
			inform(loc, "autoslab: yes maxsize "
				HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE
				" %qE %s",
				size_val, size_val, flags, lhs_type_name_ptr);
		return;
	}

	if (lhs_type != NULL_TREE) {
		tree lhs_size;

		lhs_size = TYPE_SIZE_UNIT(lhs_type);
		if (lhs_size == NULL_TREE) {
			error_at(loc, "autoslab: no nosize %qE %s", flags, lhs_type_name_ptr);
			return;
		}

		gcc_assert(TREE_CODE(lhs_size) == INTEGER_CST);
		lhs_size_val = TREE_INT_CST_LOW(lhs_size);
		lhs_align_val = TYPE_ALIGN_UNIT(lhs_type);
	} else {
		lhs_size_val = size_val;
		lhs_align_val = 0;
	}

	if (!TREE_CONSTANT(flags)) {
		if (verbose)
			inform(loc, "autoslab: no varflags "
				HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
				HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE
				"%qE %s",
				size_val, size_val, lhs_size_val, lhs_size_val, flags, lhs_type_name_ptr);
		gcc_assert(TREE_CODE(flags) == SSA_NAME);
		return;
	}

	// bail out on __GFP_DMA32/__GFP_HIGHMEM/__GFP_MOVABLE
	flags_val = TREE_INT_CST_LOW(flags);
	if (flags_val & (tree_to_uhwi(__GFP_DMA32) | tree_to_uhwi(__GFP_HIGHMEM) | tree_to_uhwi(__GFP_MOVABLE))) {
		if (verbose)
			inform(loc, "autoslab: no dmaflags "
				HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
				HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
				HOST_WIDE_INT_PRINT_HEX_PURE " %s",
				size_val, size_val, lhs_size_val, lhs_size_val, TREE_INT_CST_LOW(flags), lhs_type_name_ptr);
		return;
	}

	gcc_assert(count_arg == NULL_TREE);

#define GEN_AUTOSLAB_NAME(kind,sep1,sep2) autoslab_##kind##sep1##name_len = asprintf(&autoslab_##kind##sep1##name, \
	"autoslab." #kind #sep2 "%s.%s.%u_%u_%s_" HOST_WIDE_INT_PRINT_UNSIGNED "_" HOST_WIDE_INT_PRINT_UNSIGNED \
	"_%s_%s_%lu_" HOST_WIDE_INT_PRINT_UNSIGNED, \
	TREE_STRING_POINTER(KBUILD_MODNAME), \
	TREE_STRING_POINTER(AUTOSLAB_BASENAME), \
	LOCATION_LINE(loc), \
	LOCATION_COLUMN(loc), \
	lhs_type_name_ptr, \
	size_val, \
	lhs_align_val, \
	flags_val & tree_to_uhwi(__GFP_DMA) ? "D" : "d", \
	size_val ? "U" : "u", \
	0UL, \
	size_val);

GEN_AUTOSLAB_NAME(cache,_,.)
GEN_AUTOSLAB_NAME(init,_,.)
GEN_AUTOSLAB_NAME(,,)
#undef GEN_AUTOSLAB_NAME

	gcc_assert(autoslab_cache_name_len != -1);
	gcc_assert(autoslab_init_name_len != -1);
	gcc_assert(autoslab_name_len != -1);

	// handle the allocation of a single object
	kmem_cache_alloc = kmem_cache_decls[kmem_call].decl;
	if (kmem_cache_alloc == NULL_TREE) {
		error_at(loc, "autoslab did not find %qs to replace %qD", kmem_cache_decls[kmem_call].name, gimple_call_fndecl(kalloc_call_stmt));
		return;
	}

	autoslab_cache_id = get_identifier(autoslab_cache_name);
	var = varpool_get_for_asmname(autoslab_cache_id);
	if (var)
		kmem_cache = NODE_DECL(var);
	else {
		// build:
		// char name[] __initdata = "...";
		// struct autoslab_info i __initdata = &(struct autoslab_info){&name, size, align, flags};
		// struct autoslab_info *pi __initdata = &i;
		tree field, value, ctor, autoslab_info;
#if BUILDING_GCC_VERSION <= 4007
		VEC(constructor_elt, gc) *ctor_values = VEC_alloc(constructor_elt, gc, 5);
#else
		vec <constructor_elt, va_gc> *ctor_values = NULL;
		vec_alloc(ctor_values, 5);
#endif

		// construct the initializer for autoslab_info
		ctor = build_const_char_string(autoslab_name_len + 1, autoslab_name);

		// .name
		gcc_assert(autoslab_info_type);
		field = TYPE_FIELDS(autoslab_info_type);
		value = autoslab_build_var(loc, get_identifier(autoslab_name), TREE_TYPE(ctor), ctor, NULL);
		value = build1_loc(loc, ADDR_EXPR, TREE_TYPE(field), value);
		CONSTRUCTOR_APPEND_ELT(ctor_values, field, value);

		// .size
		field = TREE_CHAIN(field);
		value = fold_binary_to_constant(MAX_EXPR, TREE_TYPE(field), size, build_int_cst(TREE_TYPE(field), 8));
		gcc_assert(value);
		CONSTRUCTOR_APPEND_ELT(ctor_values, field, value);

		// .flags
		field = TREE_CHAIN(field);
		if (flags_val & tree_to_uhwi(__GFP_DMA))
			value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_CACHE_DMA);
		else
			value = fold_convert_loc(loc, TREE_TYPE(field), integer_zero_node);
		CONSTRUCTOR_APPEND_ELT(ctor_values, field, value);

		// .useroffset
		field = TREE_CHAIN(field);
		value = build_int_cst(TREE_TYPE(field), 0);
		CONSTRUCTOR_APPEND_ELT(ctor_values, field, value);

		// .usersize
		field = TREE_CHAIN(field);
		value = build_int_cst(TREE_TYPE(field), size_val);
		CONSTRUCTOR_APPEND_ELT(ctor_values, field, value);

		ctor = build_constructor(autoslab_info_type, ctor_values);

		TREE_STATIC(ctor) = 1;
		TREE_PUBLIC(ctor) = 0;
		TREE_CONSTANT(ctor) = 1;
		TREE_USED(ctor) = 1;

		autoslab_info = autoslab_build_var(loc, get_identifier(autoslab_init_name), autoslab_info_type, ctor, ".init.data.autoslab");

		gcc_assert(TYPE_POINTER_TO(autoslab_info_type));
		gcc_assert(POINTER_TYPE_P(TYPE_POINTER_TO(autoslab_info_type)));
		ctor = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_info_type), autoslab_info);
		kmem_cache = autoslab_build_var(loc, autoslab_cache_id, TREE_VALUE(TYPE_ARG_TYPES(TREE_TYPE(kmem_cache_alloc))), ctor, ".data.autoslab");
	}

	autoslab = create_tmp_var(TREE_TYPE(kmem_cache), autoslab_cache_name);
	add_referenced_var(autoslab);
	autoslab = make_ssa_name(autoslab, NULL);

	free(autoslab_name);
	free(autoslab_init_name);
	free(autoslab_cache_name);

	gimple_stmt_ptr assign = gimple_build_assign(autoslab, kmem_cache);
	gimple_set_location(assign, loc);
	gimple_set_block(assign, gimple_block(kalloc_call_stmt));
	gsi_insert_before(gsi, assign, GSI_SAME_STMT);
	update_stmt(assign);

	switch (kmem_call) {
	default:
		gcc_unreachable();

	case KMEM_CACHE_ALLOC:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 2, autoslab, gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(flags_arg) - 1)));
		break;

	case KMEM_CACHE_ALLOC_NODE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 3, autoslab, gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(flags_arg) - 1), gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(flags_arg))));
		break;
	}

	gimple_set_location(kmem_call_stmt, loc);
	gimple_set_block(kmem_call_stmt, gimple_block(kalloc_call_stmt));
	gimple_call_set_lhs(kmem_call_stmt, gimple_call_lhs(kalloc_call_stmt));
	gsi_replace(gsi, kmem_call_stmt, true);
	update_stmt(kmem_call_stmt);

	if (verbose)
		inform(loc, "autoslab: yes constsize "
			HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
			HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
			HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
			HOST_WIDE_INT_PRINT_HEX_PURE " "
			"%lx/%lx "
			HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
			"%s",
			size_val/lhs_size_val, size_val/lhs_size_val, size_val, size_val,
			lhs_size_val, lhs_size_val, TREE_INT_CST_LOW(flags),
			0UL, 0UL, size_val, size_val, lhs_type_name_ptr);
	gcc_assert(count_arg == 0);
}

static unsigned int autoslab_execute(void)
{
	basic_block bb;

	// 1. loop through BBs and GIMPLE statements
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			tree fndecl, count_arg, size_arg, flags_arg;
			gimple_stmt_ptr stmt;
			gcall *call_stmt;
			unsigned int kalloc_func;

			// 2. is it a call ...
			stmt = gsi_stmt(gsi);
			if (!is_gimple_call(stmt))
				continue;
			call_stmt = as_a_gcall(stmt);
			fndecl = gimple_call_fndecl(call_stmt);
			if (fndecl == NULL_TREE)
				continue;
			if (TREE_CODE(fndecl) != FUNCTION_DECL)
				continue;

			// 3. ... to k*alloc
			kalloc_func = is_kalloc(DECL_NAME_POINTER(fndecl));
			if (kalloc_func >= ARRAY_SIZE(kalloc_functions))
				continue;

			// 4. the alloc/malloc/alloc_size/alloc_flags attrs must be present on the decl/type
			gcc_assert(lookup_attribute("alloc", DECL_ATTRIBUTES(fndecl)));
			gcc_assert(lookup_attribute("malloc", DECL_ATTRIBUTES(fndecl)));
			autoslab_get_alloc_attrs(fndecl, &count_arg, &size_arg, &flags_arg, 0);
			gcc_assert(flags_arg);

			// 5. try to convert the allocation into a type specific slab allocation
			autoslab_convert_call(&gsi, call_stmt, kalloc_functions[kalloc_func].kmem_name, count_arg, size_arg, flags_arg);
		}
	}

	return 0;
}

static bool autoslab_gate(void)
{
	const char *section;

	section = get_decl_section_name(current_function_decl);
	if (section && !strncmp(section, ".init", 5))
		return false;

	if (lookup_attribute("alloc", DECL_ATTRIBUTES(current_function_decl))) {
		return false;
	}

	return true;
}

#define PASS_NAME autoslab
#define PROPERTIES_REQUIRED PROP_cfg
#if BUILDING_GCC_VERSION >= 8000 && BUILDING_GCC_VERSION < 9000
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_ggc_collect |                             TODO_verify_flow | TODO_update_address_taken
#else
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_ggc_collect | TODO_rebuild_cgraph_edges | TODO_verify_flow | TODO_update_address_taken
#endif
#include "gcc-generate-gimple-pass.h"
