/*
 * Copyright 2011 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2011-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * This gcc plugin implements a defense against memory corruption attacks by preventing unwanted
 * modifications of protected memory objects.
 *
 * This plugin is only concerned with objects with static storage duration, i.e., global variables whose
 * address is fixed at runtime (dynamically allocated objects are handled by KERNSEAL).
 *
 * Homepage:
 * http://www.grsecurity.net/~ephox/const_plugin/
 *
 * Usage:
 * $ gcc -I`gcc -print-file-name=plugin`/include -fPIC -shared -O2 -o constify_plugin.so constify_plugin.c
 * $ gcc -fplugin=constify_plugin.so test.c -O2
 *
 * TODO:
 * - add no_const to constified field if they're embedded in a non-constifiable type
 */

#include "gcc-common.h"

// unused C type flag in all versions 4.5-10
#define TYPE_CONSTIFY_VISITED(TYPE) TYPE_LANG_FLAG_4(TYPE)

__visible int plugin_is_GPL_compatible;

static GTY(()) tree pax_open_kernel_decl;
static bool pax_open_kernel_used_orig;
static bool pax_open_kernel_preserved_orig;
static GTY(()) tree pax_close_kernel_decl;
static bool pax_close_kernel_used_orig;
static bool pax_close_kernel_preserved_orig;

static bool enabled = true;
static bool verbose = false;

static struct plugin_info const_plugin_info = {
	.version	= "201705142240",
	.help		= "disable\tturn off constification\n",
};

static struct {
	const char *name;
	const char *asm_op;
} const_sections[] = {
	{".init.rodata",		"\t.section\t.init.rodata,\"a\""},
	{".ref.rodata",			"\t.section\t.ref.rodata,\"a\""},
	{".devinit.rodata",		"\t.section\t.devinit.rodata,\"a\""},
	{".devexit.rodata",		"\t.section\t.devexit.rodata,\"a\""},
	{".cpuinit.rodata",		"\t.section\t.cpuinit.rodata,\"a\""},
	{".cpuexit.rodata",		"\t.section\t.cpuexit.rodata,\"a\""},
	{".meminit.rodata",		"\t.section\t.meminit.rodata,\"a\""},
	{".memexit.rodata",		"\t.section\t.memexit.rodata,\"a\""},
	{".data.once",			"\t.section\t.data.once,\"a\""},
	{".data..read_only",		"\t.section\t.data..read_only,\"a\""},
	{".data..percpu..read_only",	"\t.section\t.data..percpu..read_only,\"a\""},
	{".pci_fixup_early",		"\t.section\t.pci_fixup_early,\"a\""},
	{".pci_fixup_header",		"\t.section\t.pci_fixup_header,\"a\""},
	{".pci_fixup_final",		"\t.section\t.pci_fixup_final,\"a\""},
	{".pci_fixup_enable",		"\t.section\t.pci_fixup_enable,\"a\""},
	{".pci_fixup_resume",		"\t.section\t.pci_fixup_resume,\"a\""},
	{".pci_fixup_resume_early",	"\t.section\t.pci_fixup_resume_early,\"a\""},
	{".pci_fixup_suspend",		"\t.section\t.pci_fixup_suspend,\"a\""},
	{".pci_fixup_suspend_late",	"\t.section\t.pci_fixup_suspend_late,\"a\""},
	{"__param",			"\t.section\t__param,\"a\""},
};

static struct {
	const char *writable_section;
	const char *readonly_section;
} constified_sections[] = {
	{ ".init.data",		".init.rodata"},
	{ ".ref.data",		".ref.rodata"},
	{ ".devinit.data",	".devinit.rodata"},
	{ ".devexit.data",	".devexit.rodata"},
	{ ".cpuinit.data",	".cpuinit.rodata"},
	{ ".cpuexit.data",	".cpuexit.rodata"},
	{ ".meminit.data",	".meminit.rodata"},
	{ ".memexit.data",	".memexit.rodata"},
};

typedef struct {
	bool has_fptr_field;
	bool has_non_integral_field;
	bool has_writable_field;
	bool has_do_const_field;
	bool has_mutable_const_field;
	bool has_no_const_field;
	bool verbose;
} constify_info;

static const_tree get_inner_type(const_tree decl)
{
	return strip_array_types(TREE_TYPE(decl));
}

/*
 * determine whether the given structure type meets the requirements for automatic constification,
 * including the constification attributes on nested structure types
 */
static void constifiable(const_tree node, constify_info *cinfo)
{
	const_tree field;

	gcc_assert(TREE_CODE(node) == RECORD_TYPE || TREE_CODE(node) == UNION_TYPE);

	// e.g., pointer to structure fields while still constructing the structure type
	if (TYPE_FIELDS(node) == NULL_TREE)
		return;

	for (field = TYPE_FIELDS(node); field; field = TREE_CHAIN(field)) {
		const_tree type = get_inner_type(field);
		enum tree_code code = TREE_CODE(type);

		if (node == type)
			continue;

		// this is not something the frontend can produce but randomize_layout does
		if (type == void_type_node)
			continue;

		if (FUNCTION_POINTER_TYPE_P(type))
			cinfo->has_fptr_field = true;
		else if (code == RECORD_TYPE || code == UNION_TYPE) {
			if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type))) {
				cinfo->has_do_const_field = true;
				if (cinfo->verbose)
					inform(DECL_SOURCE_LOCATION(field), "type %qT has field %qD with 'do_const' attribute", node, field);
			} else if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
				cinfo->has_mutable_const_field = true;
				cinfo->has_writable_field = true;
			} else if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type))) {
				cinfo->has_no_const_field = true;
				cinfo->has_writable_field = true;
			} else
				constifiable(type, cinfo);
		} else {
			if (!TREE_READONLY(field))
				cinfo->has_writable_field = true;
			if (!INTEGRAL_TYPE_P(type))
				cinfo->has_non_integral_field = true;
		}
	}
}

static bool check_attr_usage(tree type, tree name, const char *incompat_attr1, const char *incompat_attr2)
{
	if (!TYPE_P(type)) {
		error("%qE attribute applies to types only (%qD)", name, type);
		return false;
	}

	if (TREE_CODE(type) != RECORD_TYPE && TREE_CODE(type) != UNION_TYPE && TREE_CODE(type) != ARRAY_TYPE) {
		error("%qE attribute used on %qT applies to struct and union types only", name, type);
		return false;
	}

	if (lookup_attribute(IDENTIFIER_POINTER(name), TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is already applied to the type", name, type);
		return false;
	}

	if (lookup_attribute(incompat_attr1, TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is incompatible with '%s'", name, type, incompat_attr1);
		return false;
	}

	if (incompat_attr2 && lookup_attribute(incompat_attr2, TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is incompatible with '%s'", name, type, incompat_attr2);
		return false;
	}

	return true;
}

/*
 * the no_const attribute on a type ensures that the type won't be automatically constified.
 */
static tree handle_no_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	tree *type;

	*no_add_attrs = true;
	if (TREE_CODE(*node) == FUNCTION_DECL) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to functions (%qF)", name, *node);
		return NULL_TREE;
	}

	if (TREE_CODE(*node) == PARM_DECL) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to function parameters (%qD)", name, *node);
		return NULL_TREE;
	}

	if (VAR_P(*node)) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to variables (%qD)", name, *node);
		return NULL_TREE;
	}

	if (TYPE_P(*node)) {
		type = node;
		gcc_assert(flags & (int)ATTR_FLAG_TYPE_IN_PLACE);
	} else {
		if (TREE_CODE(*node) != TYPE_DECL) {
			error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to %qD (%qT)", name, *node, TREE_TYPE(*node));
			return NULL_TREE;
		}
		gcc_assert(!(flags & (int)ATTR_FLAG_TYPE_IN_PLACE));
		type = &TREE_TYPE(*node);
	}

	if (TYPE_P(*node)) {
		if (check_attr_usage(*type, name, "do_const", "mutable_const"))
			*no_add_attrs = false;
		return NULL_TREE;
	}

	if (!check_attr_usage(*type, name, "do_const", NULL))
		return NULL_TREE;

	if (!TYPE_P(*node)) {
		*type = build_variant_type_copy(*type);
		TYPE_ATTRIBUTES(*type) = tree_cons(name, NULL_TREE, TYPE_ATTRIBUTES(*type));
	}

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(*type))) {
		*no_add_attrs = false;
		return NULL_TREE;
	}

	if (enabled && TYPE_FIELDS(*type))
		error("1 %qE attribute used on type %qT that is not constified", name, *type);

	return NULL_TREE;
}

static void constify_type(tree type)
{
	tree t;

	gcc_assert(type == TYPE_MAIN_VARIANT(type));
	gcc_assert(type == TYPE_CANONICAL(type));
	TYPE_READONLY(type) = 1;
	C_TYPE_FIELDS_READONLY(type) = 1;
	TYPE_CONSTIFY_VISITED(type) = 1;

	gcc_assert(type == TYPE_CANONICAL(type));

	// fix up any already existing equivalent constified types
	for (t = TYPE_MAIN_VARIANT(type); t; t = TYPE_NEXT_VARIANT(t)) {
		tree pointer_to;

		if (t == type)
			continue;

		if (TYPE_QUALS(t) != TYPE_QUALS(type))
			continue;

		TYPE_CANONICAL(t) = TYPE_CANONICAL(type);

		if (!TYPE_POINTER_TO(t)) {
			gcc_assert(!TYPE_POINTER_TO(type));
			continue;
		}

		if (TYPE_POINTER_TO(type))
			continue;

		for (pointer_to = TYPE_POINTER_TO(t); pointer_to; pointer_to = TYPE_NEXT_PTR_TO(pointer_to))
			TREE_TYPE(pointer_to) = type;

		TYPE_POINTER_TO(type) = TYPE_POINTER_TO(t);
	}
}

/*
 * the do_const attribute on a type ensures that the type will be forcibly constified (as if
 * the type declaration had been replaced with a const typedef).
 */
static tree handle_do_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	*no_add_attrs = true;

	if (!check_attr_usage(*node, name, "no_const", "mutable_const"))
		return NULL_TREE;

	*no_add_attrs = false;
	return NULL_TREE;
}

/*
 * the mutable_const attribute on a type ensures that all objects of the type with static storage
 * duration will be forcibly constified.
 */
static tree handle_mutable_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	*no_add_attrs = true;

	if (!check_attr_usage(*node, name, "no_const", "do_const"))
		return NULL_TREE;

	*no_add_attrs = false;
	return NULL_TREE;
}

static struct attribute_spec no_const_attr = ATTRIBUTE_SPEC_INIT("no_const", 0, 0, false, false, false, handle_no_const_attribute, true, NULL);
static struct attribute_spec do_const_attr = ATTRIBUTE_SPEC_INIT("do_const", 0, 0, false, true, false, handle_do_const_attribute, true, NULL);
static struct attribute_spec mutable_const_attr = ATTRIBUTE_SPEC_INIT("mutable_const", 0, 0, false, true, false, handle_mutable_const_attribute, true, NULL);

static void register_attributes(void *event_data __unused, void *data __unused)
{
	register_attribute(&no_const_attr);
	register_attribute(&do_const_attr);
	register_attribute(&mutable_const_attr);
}

static void finish_type(void *event_data, void *data __unused)
{
	tree type = (tree)event_data;
	constify_info cinfo = {
		.has_fptr_field = false,
		.has_non_integral_field = false,
		.has_writable_field = false,
		.has_do_const_field = false,
		.has_mutable_const_field = false,
		.has_no_const_field = false,
		.verbose = false,
	};

	if (type == NULL_TREE || type == error_mark_node)
		return;

#if BUILDING_GCC_VERSION >= 5000
	if (TREE_CODE(type) == ENUMERAL_TYPE)
		return;
#endif

	if (TYPE_FIELDS(type) == NULL_TREE || TYPE_CONSTIFY_VISITED(type))
		return;

	constifiable(type, &cinfo);

	// handle mutable_const: don't allow mixing with do_const fields
	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
		if (cinfo.has_do_const_field)
			error("'mutable_const' attribute used on type %qT that is partially constified", type);
		else {
			TYPE_CONSTIFY_VISITED(type) = 1;
		}
		return;
	}

	// handle no_const: don't allow on types that wouldn't be constified anyway
	if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type))) {
		if ((cinfo.has_fptr_field && !cinfo.has_writable_field) || cinfo.has_do_const_field || cinfo.has_mutable_const_field) {
			TYPE_CONSTIFY_VISITED(type) = 1;
		} else
			error("'no_const' attribute used on type %qT that is not constified", type);
		return;
	}

	// handle do_const: don't allow on types that are eligible for automatic constification, even if they lack fptr fields
	if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type))) {
		if (!cinfo.has_writable_field) {
			error("'do_const' attribute used on type %qT that is%sautomatically constified", type, cinfo.has_fptr_field ? " " : " not ");
			return;
		}
		constify_type(type);
		return;
	}

	// propagate mutable_const from fields
	if (!cinfo.has_non_integral_field && !cinfo.has_no_const_field && !cinfo.has_do_const_field && cinfo.has_mutable_const_field) {
		TYPE_CONSTIFY_VISITED(type) = 1;
		add_type_attr(type, "mutable_const", NULL_TREE);
		return;
	}

	// automatic constification of types with fptrs and immutable fields
	if ((cinfo.has_fptr_field || cinfo.has_do_const_field) && !cinfo.has_writable_field) {
		constify_type(type);
		add_type_attr(type, "do_const", NULL_TREE);
		return;
	}

	if (cinfo.has_do_const_field) {
		error("non-constifiable type %qT has field with 'do_const' attribute", type);
		cinfo.has_fptr_field = false,
		cinfo.has_non_integral_field = false,
		cinfo.has_writable_field = false,
		cinfo.has_do_const_field = false,
		cinfo.has_mutable_const_field = false,
		cinfo.has_no_const_field = false,
		cinfo.verbose = true;
		constifiable(type, &cinfo);
	}
	TYPE_CONSTIFY_VISITED(type) = 1;
}

static bool is_const_section(const char *name)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(const_sections); i++)
		if (!strcmp(const_sections[i].name, name))
			return true;

	return false;
}

static bool is_constified_var(varpool_node_ptr node)
{
	const_tree var = NODE_DECL(node);
	const_tree type = get_inner_type(var);
	const char *name;

	if (node->alias)
		return false;

	if (DECL_EXTERNAL(var))
		return false;

	if (TREE_CODE(type) != RECORD_TYPE && TREE_CODE(type) != UNION_TYPE)
		return false;

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type)))
		return true;

	name = get_decl_section_name(var);
	if (name && is_const_section(name))
		return false;

	if (!TYPE_READONLY(type) || !C_TYPE_FIELDS_READONLY(type))
		return false;

	return TYPE_CONSTIFY_VISITED(type);
}

static void handle_section_mismatch(varpool_node_ptr node)
{
	const_tree section;
	size_t i;
	const char *name;
	tree var;

	var = NODE_DECL(node);
	name = get_decl_section_name(var);
	section = lookup_attribute("section", DECL_ATTRIBUTES(var));

	// check that the two ways of storing section info are consistently set/unset
	if (!section != !name) {
		fprintf(stderr, "DECL_SECTION: [%s] != section attr: [%s] ", name ? name : "", section ? TREE_STRING_POINTER(section) : "");
		dump_varpool_node(stderr, node);
		debug_tree(var);
		gcc_unreachable();
	}

	// the implicit default section is .data
	if (!name) {
		const_tree type;

		type = get_inner_type(var);
		if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type)))
			return;
		if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
//			inform(DECL_SOURCE_LOCATION(var), "constified mutable variable %qD moved into section .data..read_only", var);
			set_decl_section_name(var, ".data..read_only");
		} else if (!TYPE_CONSTIFY_VISITED(type)) {
			dump_varpool_node(stderr, node);
			debug_tree(var);
			gcc_unreachable();
		}
		return;
	}

//	gcc_assert(!TREE_CHAIN(section));
	gcc_assert(TREE_VALUE(section));

	section = TREE_VALUE(TREE_VALUE(section));
	gcc_assert(!strcmp(TREE_STRING_POINTER(section), name));

	if (is_const_section(name))
		return;

	for (i = 0; i < ARRAY_SIZE(constified_sections); i++) {
		if (strcmp(constified_sections[i].writable_section, name))
			continue;

//		inform(DECL_SOURCE_LOCATION(var), "constified mutable variable %qD moved into section %s", var, constified_sections[i].readonly_section);
		set_decl_section_name(var, constified_sections[i].readonly_section);
		return;
	}

	error_at(DECL_SOURCE_LOCATION(var), "constified variable %qD placed into writable section %s", var, name);
}

/*
 * this works around a gcc bug/feature where uninitialized globals
 * are moved into the .bss section regardless of any constification
 * see gcc/varasm.c:bss_initializer_p()
 */
static void fix_initializer(tree var)
{
	tree type = TREE_TYPE(var);

	if (DECL_INITIAL(var))
		return;

	DECL_INITIAL(var) = build_constructor(type, NULL);
}

static void handle_global_variables(void *event_data __unused, void *data __unused)
{
	varpool_node_ptr node;

	FOR_EACH_VARIABLE(node) {
		if (!is_constified_var(node))
			continue;

		handle_section_mismatch(node);
		fix_initializer(NODE_DECL(node));
	}
}

/*
 * the TU must provide the open/close wrappers. look them up by name early enough and mark them as used
 * otherwise gcc would delete them if the TU itself doesn't use them and before the instrumentation needs them
 */
static void find_open_close_functions(void *event_data __unused, void *data __unused)
{
	tree fndecl = (tree)event_data;
	const char *asmname;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	if (DECL_ARTIFICIAL(fndecl))
		return;

	if (DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl)
		return;

	asmname = DECL_NAME_POINTER(fndecl);
	gcc_assert(asmname[0]);

	if (DECL_NAME_LENGTH(fndecl) == sizeof("pax_open_kernel") - 1 &&
	    id_equal(DECL_NAME(fndecl), "pax_open_kernel")) {
		gcc_assert(!pax_open_kernel_decl);
		pax_open_kernel_decl = fndecl;
		pax_open_kernel_used_orig = TREE_USED(fndecl);
		TREE_USED(fndecl) = 1;
		pax_open_kernel_preserved_orig = DECL_PRESERVE_P(fndecl);
		DECL_PRESERVE_P(fndecl) = 1;
		return;
	}

	if (DECL_NAME_LENGTH(fndecl) == sizeof("pax_close_kernel") - 1 &&
	    id_equal(DECL_NAME(fndecl), "pax_close_kernel")) {
		gcc_assert(!pax_close_kernel_decl);
		pax_close_kernel_decl = fndecl;
		pax_close_kernel_used_orig = TREE_USED(fndecl);
		TREE_USED(fndecl) = 1;
		pax_close_kernel_preserved_orig = DECL_PRESERVE_P(fndecl);
		DECL_PRESERVE_P(fndecl) = 1;
		return;
	}
}

/*
 * undo the forced used marks on the open/close wrappers
 * the instrumentation no longer needs them as we're past inlining here
 */
static void unmark_open_close_functions(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr node;

	if (!pax_open_kernel_decl) {
		gcc_assert(!pax_close_kernel_decl);
		return;
	}
	gcc_assert(pax_close_kernel_decl);

	TREE_USED(pax_open_kernel_decl) = pax_open_kernel_used_orig;
	DECL_PRESERVE_P(pax_open_kernel_decl) = pax_open_kernel_preserved_orig;
	node = cgraph_get_node(pax_open_kernel_decl);
	gcc_assert(node);
#if BUILDING_GCC_VERSION <= 4007
	gcc_assert(NODE_SYMBOL(node)->needed);
	NODE_SYMBOL(node)->needed = pax_open_kernel_preserved_orig;
#else
	gcc_assert(NODE_SYMBOL(node)->force_output);
	NODE_SYMBOL(node)->force_output = pax_open_kernel_preserved_orig;
#endif

	TREE_USED(pax_close_kernel_decl) = pax_close_kernel_used_orig;
	DECL_PRESERVE_P(pax_close_kernel_decl) = pax_close_kernel_preserved_orig;
	node = cgraph_get_node(pax_close_kernel_decl);
	gcc_assert(node);
#if BUILDING_GCC_VERSION <= 4007
	gcc_assert(NODE_SYMBOL(node)->needed);
	NODE_SYMBOL(node)->needed = pax_close_kernel_preserved_orig;
#else
	gcc_assert(NODE_SYMBOL(node)->force_output);
	NODE_SYMBOL(node)->force_output = pax_close_kernel_preserved_orig;
#endif
}

static unsigned int check_local_variables_execute(void)
{
	unsigned int ret = 0;
	tree var;

	unsigned int i;

	FOR_EACH_LOCAL_DECL(cfun, i, var) {
		tree type = TREE_TYPE(var);

		gcc_assert(DECL_P(var));
		if (is_global_var(var))
			continue;

		if (TREE_CODE(type) != RECORD_TYPE && TREE_CODE(type) != UNION_TYPE)
			continue;

		if (!TYPE_READONLY(type) || !C_TYPE_FIELDS_READONLY(type))
			continue;

		if (!TYPE_CONSTIFY_VISITED(type))
			continue;

		error_at(DECL_SOURCE_LOCATION(var), "constified variable %qE cannot be local", var);
		ret = 1;
	}
	return ret;
}

#define PASS_NAME check_local_variables
#define NO_GATE
#include "gcc-generate-gimple-pass.h"

static bool is_write(gimple_stmt_ptr stmt)
{
	if (!is_gimple_assign(stmt))
		return false;

	if (gimple_clobber_p(stmt))
		return false;

	return gimple_store_p(stmt);
}

static bool is_mutable_write(gimple_stmt_ptr stmt)
{
	tree lhs, ref, offset, reftype;
	poly_int64_pod bitsize, bitpos;
	enum machine_mode mode;
	int unsignedp, reversep, volatilep;
	const char *name;

	if (!is_write(stmt))
		return false;

	lhs = gimple_get_lhs(stmt);

	// simple parameter write
	if (TREE_CODE(lhs) == PARM_DECL)
		return false;

	// simple return value write
	if (TREE_CODE(lhs) == RESULT_DECL)
		return false;

	ref = get_inner_reference(lhs, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
	reftype = TREE_TYPE(ref);

	switch (TREE_CODE(ref)) {
	default:
		print_gimple_stmt(stderr, stmt, 0, TDF_LINENO);
		debug_tree(lhs);
		debug_tree(ref);
		gcc_unreachable();

	case VAR_DECL:
		if (!is_global_var(ref))
			return false;

		if (lookup_attribute("no_const", TYPE_ATTRIBUTES(reftype)))
			return false;

		if (lookup_attribute("do_const", TYPE_ATTRIBUTES(reftype)) || lookup_attribute("mutable_const", TYPE_ATTRIBUTES(reftype)))
			return true;

		name = get_decl_section_name(ref);
		return name && is_const_section(name);

	case INDIRECT_REF:
	case TARGET_MEM_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
		return false;

	case PARM_DECL:
		if (!offset)
			return false;
		gcc_unreachable();
	case RESULT_DECL:
		return false;
	}
}

static bool is_unsafe_write(gimple_stmt_ptr stmt)
{
	const_tree lhs;

	if (is_gimple_debug(stmt))
		return false;

	if (!is_gimple_assign(stmt))
		return true;

	if (!is_write(stmt))
		return false;

	lhs = gimple_get_lhs(stmt);

	switch (TREE_CODE(lhs)) {
	default:
		return true;

	// simple parameter write
	case PARM_DECL:
		return false;

	// simple return value write
	case RESULT_DECL:
		return false;

	case VAR_DECL:
		return !INTEGRAL_TYPE_P(TREE_TYPE(lhs));
	}
}

enum placement {before, after};

static void constify_insert_call(tree callee, gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	gimple_stmt_ptr stmt;
	gcall *gcallee;
	cgraph_node_ptr node;
	int frequency;
	basic_block bb;

	// insert call to callee
	stmt = gimple_build_call(callee, 0);
	gimple_set_location(stmt, gimple_location(reference));
	gimple_set_block(stmt, gimple_block(reference));
	gcallee = as_a_gcall(stmt);
	if (placement == before)
		gsi_insert_before(gsi, gcallee, GSI_SAME_STMT);
	else
		gsi_insert_after(gsi, gcallee, GSI_CONTINUE_LINKING);

	// update the cgraph
	bb = gimple_bb(gcallee);
	node = cgraph_get_node(callee);
	gcc_assert(node);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), node, gcallee, bb->count, frequency, bb->loop_depth);
}

static void pax_open_kernel(gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	if (!pax_open_kernel_decl) {
		error_at(gimple_location(reference), "pax_open_kernel must be declared");
		return;
	}

	if (verbose)
		inform(gimple_location(gsi_stmt(*gsi)), "writable access to constified region begins %s", placement == before ? "before" : "after");

	constify_insert_call(pax_open_kernel_decl, gsi, reference, placement);
}

static void pax_close_kernel(gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	if (!pax_close_kernel_decl) {
		error_at(gimple_location(reference), "pax_close_kernel must be declared");
		return;
	}

	if (verbose) {
		if (gsi_end_p(*gsi))
			inform(gimple_location(reference), "writable access to constified region ends %s (approximately)", placement == before ? "before" : "after");
		else
			inform(gimple_location(gsi_stmt(*gsi)), "writable access to constified region ends %s", placement == before ? "before" : "after");
	}

	constify_insert_call(pax_close_kernel_decl, gsi, reference, placement);
}

static unsigned int instrument_mutable_writes_execute(void)
{
	basic_block bb;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_ptr stmt;
		gimple_stmt_iterator gsi;
		gimple_stmt_iterator gsi_open = gsi_none();
		gimple_stmt_iterator gsi_close = gsi_none();
		bool has_unsafe_write = false;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			stmt = gsi_stmt(gsi);

			// record the beginning/end of a sequence of mutable write stmts
			if (is_mutable_write(stmt)) {
				if (gsi_end_p(gsi_open))
					gsi_open = gsi;
				gsi_close = gsi;
				continue;
			}

			// an unsafe write forces instrumentation...
			if (!is_unsafe_write(stmt))
				continue;

			gcc_assert(gsi_end_p(gsi_open) == gsi_end_p(gsi_close));

			// ...unless we haven't found a mutable write yet
			// record the fact that this block is not eligible for optimization
			if (gsi_end_p(gsi_open)) {
				has_unsafe_write = true;
				continue;
			}

			// emit instrumentation and reset the state
			pax_open_kernel(&gsi_open, stmt, before);
			pax_close_kernel(&gsi_close, stmt, after);

			gsi_open = gsi_none();
			gsi_close = gsi_none();
		}

		gcc_assert(gsi_end_p(gsi_open) == gsi_end_p(gsi_close));

		if (gsi_end_p(gsi_open))
			continue;

		// instrumentation is needed but it can't be moved out of the loop due to unsafe writes
		if (has_unsafe_write) {
			pax_open_kernel(&gsi_open, gsi_stmt(gsi_open), before);
			pax_close_kernel(&gsi_close, gsi_stmt(gsi_close), after);
			continue;
		}

		// loops with no unsafe writes allow moving the instrumentation outside the loop
		// handle the simplest cases for now
		if (bb_any_loop(bb) && bb->loop_father->num_nodes == 2) {
			basic_block bb_open, bb_close;

			// place the open call into the immediate single predecessor dominator block
			bb_open = get_immediate_dominator(CDI_DOMINATORS, bb->loop_father->header);
			gcc_assert(single_succ_p(bb_open));
			gsi_open = gsi_last_nondebug_bb(bb_open);
			pax_open_kernel(&gsi_open, gsi_stmt(gsi_open), after);

			// place the close call into the immediate single successor post-dominator block
			// if there isn't one then split each loop exiting edge
			bb_close = get_immediate_dominator(CDI_POST_DOMINATORS, bb->loop_father->header);

			if (single_pred_p(bb_close)) {
				gsi_close = gsi_start_nondebug_after_labels_bb(bb_close);
				pax_close_kernel(&gsi_close, gsi_stmt(gsi_close), before);
			} else {
				edge e;
				edge_iterator ei;

				stmt = gsi_stmt(gsi_close);

				FOR_EACH_EDGE(e, ei, bb_close->preds) {
					basic_block new_close;

					if (e->src->loop_father != bb->loop_father)
						continue;

					new_close = plugin_split_edge(e);
					gcc_assert(single_pred_p(new_close));
					gsi_close = gsi_start_nondebug_after_labels_bb(new_close);
					pax_close_kernel(&gsi_close, stmt, after);
				}
			}

			continue;
		}

		// give up and just instrument
		pax_open_kernel(&gsi_open, gsi_stmt(gsi_open), before);
		pax_close_kernel(&gsi_close, gsi_stmt(gsi_close), after);
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();
	return 0;
}

#define PASS_NAME instrument_mutable_writes
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_ggc_collect | TODO_rebuild_cgraph_edges | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"

static unsigned int (*old_section_type_flags)(tree decl, const char *name, int reloc);

static unsigned int constify_section_type_flags(tree decl, const char *name, int reloc)
{
	return is_const_section(name) ? 0 : old_section_type_flags(decl, name, reloc);
}

static bool (*old_lang_hooks_init)(void);

static bool constify_lang_hooks_init(void)
{
	register_callback("constify_plugin", PLUGIN_FINISH_TYPE, finish_type, NULL);
	return old_lang_hooks_init();
}

static void constify_start_unit(void *gcc_data __unused, void *user_data __unused)
{
	old_section_type_flags = targetm.section_type_flags;
	targetm.section_type_flags = constify_section_type_flags;
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	int i;

	static const struct ggc_root_tab gt_ggc_r_gt_constify[] = {
		{
			.base = &pax_open_kernel_decl,
			.nelt = 1,
			.stride = sizeof(pax_open_kernel_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		{
			.base = &pax_close_kernel_decl,
			.nelt = 1,
			.stride = sizeof(pax_close_kernel_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		LAST_GGC_ROOT_TAB
	};

	PASS_INFO(check_local_variables, "ssa", 1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(instrument_mutable_writes, "ssa", 1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!(strcmp(argv[i].key, "disable"))) {
			enabled = false;
			continue;
		}
		if (!(strcmp(argv[i].key, "verbose"))) {
			verbose = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	if (strncmp(lang_hooks.name, "GNU C", 5) || !strncmp(lang_hooks.name, "GNU C+", 6)) {
		inform(UNKNOWN_LOCATION, G_("%s supports C only, not %s"), plugin_name, lang_hooks.name);
		enabled = false;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &const_plugin_info);
	if (enabled) {
		old_lang_hooks_init = lang_hooks.init;
		lang_hooks.init = constify_lang_hooks_init;
		register_callback(plugin_name, PLUGIN_START_UNIT, constify_start_unit, (void *)plugin_name);
		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, find_open_close_functions, NULL);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, handle_global_variables, NULL);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, unmark_open_close_functions, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &check_local_variables_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &instrument_mutable_writes_pass_info);
		register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_constify);
	}
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	return 0;
}
