/*
 * Copyright 2012-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

bool report_fptr_hash;

static bool rap_fptr_gate(void)
{
	return rap_cmodel_check();
}

static tree build_rap_hash(gimple_stmt_ptr call_stmt, tree fntype)
{
	rap_hash_t hash;

	hash = rap_lookup_imprecise_rap_hash(fntype);
	if (report_fptr_hash)
		inform(gimple_location(call_stmt), "fptr rap_hash: %x", hash.hash);
	return build_int_cst_type(rap_hash_type_node, hash.hash);
}

// extract the rap_hash stored at an offset from ptr
tree get_rap_hash(gimple_seq *stmts, location_t loc, tree fptr, HOST_WIDE_INT rap_hash_offset)
{
	gimple_stmt_ptr assign_hash;
	tree target_hash;
	tree fptr2;

	// target_hash = *(long*)((void*)fptr + rap_hash_offset)
	target_hash = create_tmp_var(rap_hash_type_node, "rap_hash");
	add_referenced_var(target_hash);
	target_hash = make_ssa_name(target_hash, NULL);
#if BUILDING_GCC_VERSION == 4005
	fptr2 = create_tmp_var(ptr_type_node, "rap_fptr2");
	fptr2 = make_ssa_name(fptr2, NULL);
	assign_hash = gimple_build_assign(fptr2, build2_loc(loc, POINTER_PLUS_EXPR, ptr_type_node, fptr, build_int_cst_type(sizetype, rap_hash_offset)));
	gimple_set_location(assign_hash, loc);
	SSA_NAME_DEF_STMT(fptr2) = assign_hash;
	gimple_seq_add_stmt(stmts, assign_hash);
	fptr = gimple_get_lhs(assign_hash);

	fptr2 = create_tmp_var(build_pointer_type(rap_hash_type_node), "rap_fptr2");
	fptr2 = make_ssa_name(fptr2, NULL);
	assign_hash = gimple_build_assign(fptr2, fold_convert_loc(loc, build_pointer_type(TREE_TYPE(target_hash)), fptr));
	gimple_set_location(assign_hash, loc);
	SSA_NAME_DEF_STMT(fptr2) = assign_hash;
	gimple_seq_add_stmt(stmts, assign_hash);
	fptr = gimple_get_lhs(assign_hash);

	assign_hash = gimple_build_assign(target_hash, build1_loc(loc, INDIRECT_REF, rap_hash_type_node, fptr));
#else
#ifdef TARGET_386
	if (!TARGET_64BIT)
		fptr2 = create_tmp_var(build_qualified_type(ptr_type_node, ENCODE_QUAL_ADDR_SPACE(ADDR_SPACE_KERNEXEC)), "rap_fptr2");
	else
#endif
		fptr2 = create_tmp_var(ptr_type_node, "rap_fptr2");
	fptr2 = make_ssa_name(fptr2, NULL);
	assign_hash = gimple_build_assign(fptr2, fptr);
	gimple_set_location(assign_hash, loc);
	SSA_NAME_DEF_STMT(fptr2) = assign_hash;
	gimple_seq_add_stmt(stmts, assign_hash);
	fptr = gimple_get_lhs(assign_hash);

	assign_hash = gimple_build_assign(target_hash, build2_loc(loc, MEM_REF, rap_hash_type_node, fptr, build_int_cst_type(build_pointer_type(rap_hash_type_node), rap_hash_offset)));
#endif
	gimple_set_location(assign_hash, loc);
	SSA_NAME_DEF_STMT(target_hash) = assign_hash;
	gimple_seq_add_stmt(stmts, assign_hash);

	return target_hash;
}

// check the function hash of the target of the fptr
static void rap_instrument_fptr(gimple_stmt_iterator *gsi)
{
	gimple_stmt_ptr check_hash, call_stmt, stmt;
	gimple_seq stmts = NULL;
	location_t loc;
	tree block, computed_hash, target_hash, fptr, fntype;
	basic_block cond_bb, join_bb, true_bb;
	edge e;
	HOST_WIDE_INT rap_hash_offset;

	call_stmt = gsi_stmt(*gsi);
	loc = gimple_location(call_stmt);
	block = gimple_block(call_stmt);
	fptr = gimple_call_fn(call_stmt);
	fntype = gimple_call_fntype(call_stmt);

	if (TREE_CODE(fntype) != FUNCTION_TYPE) {
		debug_tree(fntype);
		gcc_unreachable();
	}

	computed_hash = build_rap_hash(call_stmt, fntype);

	rap_hash_offset = UNITS_PER_WORD / 4 * sizeof(rap_hash_t);
	target_hash = get_rap_hash(&stmts, loc, fptr, -rap_hash_offset);
	gsi_insert_seq_before(gsi, stmts, GSI_SAME_STMT);

	// compare target_hash against computed function hash
	// bail out on mismatch
	check_hash = gimple_build_cond(NE_EXPR, target_hash, computed_hash, NULL_TREE, NULL_TREE);
	gimple_set_location(check_hash, loc);
	gimple_set_block(check_hash, block);
	gsi_insert_before(gsi, check_hash, GSI_NEW_STMT);

#if 1
	cond_bb = gimple_bb(gsi_stmt(*gsi));
	gcc_assert(!gsi_end_p(*gsi));
	gcc_assert(check_hash == gsi_stmt(*gsi));

	e = split_block(cond_bb, gsi_stmt(*gsi));
	gcc_assert(cond_bb == e->src);
	join_bb = e->dest;
	e->flags = EDGE_FALSE_VALUE;
	e->probability -= REG_BR_PROB_VERY_UNLIKELY;

	true_bb = create_empty_bb(EXIT_BLOCK_PTR_FOR_FN(cfun)->prev_bb);
	e = make_edge(cond_bb, true_bb, EDGE_TRUE_VALUE | EDGE_PRESERVE);
	e->probability = REG_BR_PROB_VERY_UNLIKELY;
	true_bb->count = edge_count(e);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	set_immediate_dominator(CDI_DOMINATORS, true_bb, cond_bb);
	set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

	gcc_assert(cond_bb->loop_father == join_bb->loop_father);
	add_bb_to_loop(true_bb, cond_bb->loop_father);
#else
	true_bb = insert_cond_bb(gimple_bb(gsi_stmt(*gsi)), gsi_stmt(*gsi), check_hash, REG_BR_PROB_VERY_UNLIKELY);
#endif
	loops_state_set(LOOPS_NEED_FIXUP);

	*gsi = gsi_start_bb(true_bb);

	if (rap_call_abort) {
		tree input;
#if BUILDING_GCC_VERSION <= 4007
		VEC(tree, gc) *inputs = NULL;
#else
		vec<tree, va_gc> *inputs = NULL;
#endif

		// asm volatile("rap_call_abort %0" : : "r"(fptr));
		input = build_tree_list(NULL_TREE, build_const_char_string(2, "r"));
		input = chainon(NULL_TREE, build_tree_list(input, fptr));
#if BUILDING_GCC_VERSION <= 4007
		VEC_safe_push(tree, gc, inputs, input);
#else
		vec_safe_push(inputs, input);
#endif
		stmt = gimple_build_asm_vec(rap_call_abort, inputs, NULL, NULL, NULL);
		gimple_asm_set_volatile(as_a_gasm(stmt), true);
		gimple_set_location(stmt, loc);
		gimple_set_block(stmt, block);
		gsi_insert_after(gsi, stmt, GSI_CONTINUE_LINKING);

		stmt = gimple_build_call(builtin_decl_implicit(BUILT_IN_UNREACHABLE), 0);
	} else {
		// this fake dependency is to prevent PRE from merging this BB with others of the same kind
		stmt = barrier(fptr, false);
		gimple_set_location(stmt, loc);
		gimple_set_block(stmt, block);
		gsi_insert_after(gsi, stmt, GSI_CONTINUE_LINKING);

		stmt = gimple_build_call(builtin_decl_implicit(BUILT_IN_TRAP), 0);
	}

	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	gsi_insert_after(gsi, stmt, GSI_CONTINUE_LINKING);

	*gsi = gsi_start_bb(join_bb);
}

// find all language level function pointer dereferences and verify the target function
static unsigned int rap_fptr_execute(void)
{
	basic_block bb;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	// 1. loop through BBs and GIMPLE statements
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			// gimple match: h_1 = get_fptr (); D.2709_3 = h_1 (x_2(D));
			tree fptr, fntype;
			gimple_stmt_ptr call_stmt;

			// is it a call ...
			call_stmt = gsi_stmt(gsi);
			if (!is_gimple_call(call_stmt))
				continue;

			fptr = gimple_call_fn(call_stmt);
			if (!fptr)
				continue;

			fntype = gimple_call_fntype(call_stmt);
			switch (TREE_CODE(fptr)) {
			default:
				debug_gimple_stmt(call_stmt);
				debug_tree(fptr);
				debug_tree(TREE_TYPE(fptr));
				debug_tree(fntype);
				gcc_unreachable();

			case ADDR_EXPR:
				continue;

			case SSA_NAME:
				if (SSA_NAME_VAR(fptr) == NULL_TREE)
					break;

				switch (TREE_CODE(SSA_NAME_VAR(fptr))) {
				default:
					debug_gimple_stmt(call_stmt);
					debug_tree(fptr);
					debug_tree(fntype);
					gcc_unreachable();

				case VAR_DECL:
				case PARM_DECL:
					break;
				}
				break;

			case INTEGER_CST:
			case OBJ_TYPE_REF:
				break;
			}

			// ... through a function pointer
			gcc_assert(POINTER_TYPE_P(TREE_TYPE(fptr)));
			gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);

			if (enable_type_call)
				rap_instrument_fptr(&gsi);

			bb = gsi_bb(gsi);
			gcc_assert(call_stmt == gsi_stmt(gsi));
		}
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

#define PASS_NAME rap_fptr
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_rebuild_cgraph_edges | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"
